"""Outset Python Package"""
from __future__ import annotations

__version__ = "0.0.0""-rc53-post1"

from outset.draw_outset_ import draw_outset
from outset.mark_arrow_ import mark_arrow
from outset.mark_inlaid_asterisk_ import mark_inlaid_asterisk
from outset.mark_magnifying_glass_ import mark_magnifying_glass

__all__ = [
    "draw_outset",
    "mark_arrow",
    "mark_inlaid_asterisk",
    "mark_magnifying_glass",
]



import typing

from matplotlib import axes as mpl_axes
from matplotlib import patches as mpl_patches
from matplotlib import pyplot as plt
from matplotlib import colors as mpl_colors
import numpy as np

from ._auxlib.compose_zoom_trapezoid_ import compose_zoom_trapezoid
from ._auxlib.get_vertices_extent_ import get_vertices_extent
from ._auxlib.find_intersection_ import find_intersection
from ._auxlib.make_radial_gradient_ import make_radial_gradient
from ._auxlib.minimal_perimeter_permutation_ import (
    minimal_perimeter_permutation,
)


def draw_outset(
    xlim: typing.Tuple[float, float],
    ylim: typing.Tuple[float, float],
    ax: typing.Optional[mpl_axes.Axes] = None,
    color: str = "blue",
    box_facecolor: str = "white",
    box_linewidth: float = 0.5,
    clip_on: bool = False,
    hide_outer_spines: bool = True,
    markersize: float = 15,
    zoom_linestyle: str = ":",
    zoom_linewidth: int = 2,
    stretch: float = 0.14,
) -> mpl_axes.Axes:
    """Mark the boundary of a rectangular region and annotate with a flyout
    "zoom" indication upwards and to the right.

    Parameters
    ----------
    xlim : Tuple[float, float]
        The x-limits of the rectangular boundary in the form (xmin, xmax).
    ylim : Tuple[float, float]
        The y-limits of the rectangular boundary in the form (ymin, ymax).
    ax : matplotlib.axes.Axes, optional
        The axes object on which to draw. If None, the current active axes will
        be used.
    color : str, default "blue"
        The color of the box's edge and the zoom indication's lines.
    box_facecolor : str, default "white"
        The fill color of the rectangular boundary.
    box_linewidth : float, default 0.5
        The line width of the rectangular boundary's edge.
    clip_on : bool, default False
        If True, the drawing elements are clipped to the axes bounding box.
    hide_outer_spines : bool, default True
        If True, hides the right and top spines of the axes.
    markersize : float, default 15
        The size of the marker at the intersection of the zoom indication's
        edges.
    zoom_linestyle : str, default ":"
        The line style for the zoom indication (e.g., solid, dashed, dotted).
    zoom_linewidth : int, default 2
        The line width of the zoom indication's edges.
    stretch : float, default 0.14
        How far should zoom indication stretch upwards and to the right?

    Returns
    -------
    mpl_axes.Axes
        The modified matplotlib axes object with the drawing elements added.
    """
    if ax is None:
        ax = plt.gca()

    # Outline zoom regions with rectangle
    ###########################################################################
    rect = mpl_patches.Rectangle(
        (xlim[0], ylim[0]),  # lower left corner
        xlim[1] - xlim[0],  # width
        ylim[1] - ylim[0],  # height
        linewidth=box_linewidth,
        edgecolor=color,
        facecolor=box_facecolor,
        clip_on=clip_on,
    )
    ax.add_patch(rect)

    # Draw zoom trapezoid...
    ###########################################################################
    trapezoid_vertices = compose_zoom_trapezoid(
        xlim, ylim, ax.get_xlim(), ax.get_ylim(), stretch=stretch
    )
    trapezoid_vertices = minimal_perimeter_permutation(trapezoid_vertices)
    # ... outline
    trapezoid = mpl_patches.Polygon(
        trapezoid_vertices,
        closed=True,
        facecolor="none",
        edgecolor=color,
        linestyle=zoom_linestyle,
        linewidth=zoom_linewidth,
        zorder=-3,
        clip_on=clip_on,
    )
    ax.add_patch(trapezoid)

    # ... gradient fill, clipped insidetrapezoid
    img = ax.imshow(
        make_radial_gradient(),
        extent=get_vertices_extent(trapezoid_vertices),
        interpolation="nearest",
        aspect="auto",
        cmap=mpl_colors.LinearSegmentedColormap.from_list(
            "gradient",
            ["white", color],
        ),
    )
    if not clip_on:
        img.set_clip_box(ax.bbox.shrunk(10, 10))  # grow axis clipping box
    img.set_clip_path(trapezoid)

    # Add circle and asterisk at point where trapezoid edges meet...
    ###########################################################################
    rotated_vertices = _rotate_vertices(trapezoid_vertices)
    center = find_intersection(*rotated_vertices)

    plt.plot(  # ...white underlay for asterisk
        *center,
        marker="o",
        markeredgecolor="none",
        markerfacecolor="white",
        markersize=markersize * 1.66,
        clip_on=clip_on,
    )
    plt.plot(  # ...asterisk
        *center,
        marker=(6, 2, 0),  # six-pointed asterisk
        markeredgecolor=color,
        markerfacecolor="none",
        markersize=markersize,
        clip_on=clip_on,
    )

    # Finalize
    ###########################################################################
    ax.set_axisbelow(True)  # ensure annotations above if outside bounds
    if hide_outer_spines:
        ax.spines[["right", "top"]].set_visible(False)


def _rotate_vertices(trapezoid_vertices: np.ndarray) -> np.ndarray:
    # need to roll trapezoid_vertices so that the rectangle-corner vertices
    # are in first and last positions
    assert len(trapezoid_vertices) == 4

    xs = np.array(trapezoid_vertices)[:, 0]
    ys = np.array(trapezoid_vertices)[:, 1]
    rect_upper_left_idx, rect_lower_right_idx = np.argmin(xs), np.argmin(ys)

    gap = (rect_lower_right_idx - rect_upper_left_idx) % 4
    if gap == 1:
        rotated = np.roll(trapezoid_vertices, -rect_lower_right_idx, axis=0)
    elif gap == 3:
        rotated = np.roll(trapezoid_vertices, -rect_upper_left_idx, axis=0)
    else:
        assert False, (gap, trapezoid_vertices)

    assert np.argmin(np.array(rotated)[:, 0]) in (0, 3)
    assert np.argmin(np.array(rotated)[:, 1]) in (0, 3)

    return rotated



import numpy as np


def make_radial_gradient() -> np.ndarray:
    """Generate a radial gradient intensity map.

    Gradient takes maximum intensity in upper-right corner of array.

    Returns
    -------
    np.ndarray
        A 2D NumPy array of shape (101, 101), containing the normalized radial
        gradient values ranging from 0 to 1.
    """

    x = np.linspace(-5, 0, 101)
    y = np.linspace(0, 5, 101)

    xs, ys = np.meshgrid(x, y)
    zs = -np.sqrt(xs**2 + ys**2)

    z_norm = (zs - np.min(zs.flat)) / (np.max(zs.flat) - np.min(zs.flat))
    # note: must assign to new variable or incorrect behavior occurs
    zs = np.power(z_norm, 2)  # nonlinear gradient rate
    z_norm = (zs - np.min(zs.flat)) / (np.max(zs.flat) - np.min(zs.flat))

    return z_norm



import matplotlib.pyplot as plt

from outset import mark_magnifying_glass


def test_mark_magnifying_glass_single():
    plt.clf()  # Clear the current figure
    ax = plt.gca()  # Get the current axes

    # Call the function with test values
    mark_magnifying_glass(x=0.5, y=0.5, ax=ax, color="red")

    # Save and print the path of the output image
    outpath = "/tmp/test_mark_magnifying_glass_single.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")
    plt.close()  # Close the plot after saving


def test_mark_magnifying_glass_multiple():
    plt.clf()
    # Multiple calls to the function with different parameters
    mark_magnifying_glass(x=0.3, y=0.3)
    mark_magnifying_glass(x=0.7, y=0.7, color="blue", color_accent="orange")

    # Save and print the path of the output image
    outpath = "/tmp/test_mark_magnifying_glass_multiple.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")
    plt.close()



import math

from outset._auxlib.find_intersection_ import find_intersection


def test_intersection_typical():
    # Test a typical case where two lines intersect
    assert find_intersection((0, 0), (1, 1), (0, 1), (1, 0)) == (0.5, 0.5)


def test_intersection_parallel():
    # Test a case where two lines are parallel and don't intersect
    assert math.isnan(find_intersection((0, 0), (1, 1), (0, 1), (1, 2))[0])
    assert math.isnan(find_intersection((0, 0), (1, 1), (0, 1), (1, 2))[1])


def test_intersection_collinear():
    # Test a case where two lines are collinear
    assert math.isnan(find_intersection((0, 0), (1, 1), (2, 2), (3, 3))[0])
    assert math.isnan(find_intersection((0, 0), (1, 1), (2, 2), (3, 3))[1])


def test_intersection_points_underspecified():
    # Test a case where lines are underspecified (e.g., points are identical)
    assert math.isnan(find_intersection((0, 0), (0, 0), (1, 1), (1, 1))[0])
    assert math.isnan(find_intersection((0, 0), (0, 0), (1, 1), (1, 1))[1])



import typing

import numpy as np

from .minimal_perimeter_permutation_ import minimal_perimeter_permutation


def compose_zoom_trapezoid(
    rect_xlim: typing.Tuple[float, float],
    rect_ylim: typing.Tuple[float, float],
    ax_xlim: typing.Tuple[float, float],
    ax_ylim: typing.Tuple[float, float],
    stretch: float,
):
    """Decide shape, size, and position of trapezoidal zoom effect decoration.

    Uses reflected right triangles similar to the proportions of the rectangle
    to pick the upper points for the trapezoid, relative to the upper right
    corner of the rectangle. The trapezoid is then constructed to include
    the upper left corner of the rectangle and the lower right corner of the rectangle.

    The distance of the upper vertices of the zoom annotation from the
    rectangle are controlled by scaling the similar triangles. Scaling is
    adjusted by relative size of the rectangle within the axis viewport,
    with smaller rectangles given proportionally larger stretch factors to
    prevent overlap of decorations applied to the upper vertices with the
    rectangle.

    This can be tweaked through the "stretch" parameter, with higher stretch
    factors resulting in upper vertices further from the rectangle.

    Note that a portion of the trapezoid falls behind the rectangle.

    Limits should be provided in ascending order.

    Parameters
    ----------
    rect_xlim : typing.Tuple[float, float]
        The x-axis limits (min, max) of the rectangle.
    rect_ylim : typing.Tuple[float, float]
        The y-axis limits (min, max) of the rectangle.
    ax_xlim : typing.Tuple[float, float]
        The x-axis limits (min, max) of the axis.
    ax_ylim : typing.Tuple[float, float]
        The y-axis limits (min, max) of the axis.
    stretch : float
        The stretch factor for the zoom effect.

    Returns
    -------
    list
        A list of tuples representing the sequence of vertices of the trapezoid, arranged to minimize the perimeter.

    """
    if any(
        float.__gt__(*map(float, lim))
        for lim in [rect_xlim, rect_ylim, ax_xlim, ax_ylim]
    ):
        raise ValueError("Limits must be provided as (min, max).")

    ax_width, ax_height = (
        ax_xlim[1] - ax_xlim[0],
        ax_ylim[1] - ax_ylim[0],
    )
    rect_width, rect_height = (
        rect_xlim[1] - rect_xlim[0],
        rect_ylim[1] - rect_ylim[0],
    )
    (
        (_rect_lower_left, rect_upper_right),
        (rect_upper_left, rect_lower_right),
    ) = zip(rect_xlim, rect_ylim), zip(rect_xlim, reversed(rect_ylim))

    rel_width = rect_width / ax_width
    rel_height = rect_height / ax_height
    mx = stretch / np.sqrt(rel_width**2 + rel_height**2)
    offsets = np.array([mx * rect_height, mx * rect_width])

    trapezoid_top_left = np.array(rect_upper_right) + offsets
    trapezoid_top_right = np.array(rect_upper_right) + np.flip(offsets)

    trapezoid_points = [
        rect_upper_left,
        rect_lower_right,
        trapezoid_top_right,
        trapezoid_top_left,
    ]
    # ensure no "bow-tie" self intersection
    trapezoid_sequence = minimal_perimeter_permutation(trapezoid_points)

    return trapezoid_sequence



import matplotlib.pyplot as plt

from outset import draw_outset


def test_draw_outset_one():
    plt.clf()
    draw_outset(
        xlim=(0, 1),
        ylim=(0, 2),
        color="mediumpurple",
        box_facecolor="white",
        box_linewidth=0.5,
        clip_on=False,
        hide_outer_spines=True,
        markersize=15,
        zoom_linestyle=":",
        zoom_linewidth=2,
        stretch=0.14,
    )
    outpath = "/tmp/test_draw_outset_one.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_draw_outset_several():
    _fig, ax = plt.subplots(figsize=(6, 4))

    draw_outset(
        xlim=[1, 1.25],
        ylim=(0.5, 1.5),
        ax=ax,
    )
    draw_outset(
        xlim=[2, 3.9],
        ylim=(0.5, 1.5),
        ax=ax,
    )
    draw_outset(
        xlim=[1, 2],
        ylim=(2, 3),
        ax=ax,
    )

    # Set limits and hide axes for a clean presentation
    ax.set_xlim(0, 4)
    ax.set_ylim(0, 4)

    # Display the updated plot
    outpath = "/tmp/test_draw_outset_several.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")
